@@ -17,11 +17,13 @@
 #
 # SCCS: @(#) httpd.tcl 1.16 97/07/12 14:29:54
 
-
 package provide httpd 1.1
 
 # initialize all the global data
 
+# Location of this package
+set Httpd(library) [file dirname [info script]]
+
 # HTTP/1.0 error codes (the ones we use)
 array set Httpd_Errors {
     200 {Data follows}
@@ -143,7 +145,7 @@ proc Httpd_Shutdown {} {
     set ok 1
     foreach handler $Httpd(shutdown) {
 	if [catch {eval $handler} err] {
-	    Log "Shutdown: $handler:\n$err"
+	    Log "" "Shutdown: $handler" $err
 	    set ok 0
 	}
     }
@@ -242,6 +244,10 @@ proc HttpdRead {sock} {
 			$line x data(proto) data(url) data(query) data(version)] {
 		    set data(state) mime
 		    set data(line) $line
+		    set data(uri) $data(url)
+		    if {[string length $data(query)]} {
+			append data(uri) ?$data(query)
+		    }
 		    CountHist urlhits
 		    # Limit the time allowed to serve this request
 		    catch {after cancel $data(cancel)}
@@ -283,7 +289,7 @@ proc HttpdRead {sock} {
 	        	[info exists data(mime,content-length)]} {
 		    set data(linemode) 0
 	            set data(count) $data(mime,content-length)
-	            if {$data(version) >= 1.1 && [info exists data(mime,expect]} {
+	            if {$data(version) >= 1.1 && [info exists data(mime,expect)]} {
 	                if {$data(mime,expect) == "100-continue"} {
 			    puts $sock "100 Continue HTTP/1.1\n"
 			    flush $sock
@@ -294,6 +300,7 @@ proc HttpdRead {sock} {
 		    fconfigure $sock  -translation {binary crlf}
 		} elseif {$data(proto) != "POST"}  {
 		    # Dispatch upon blank line after headers
+		    fileevent $sock readable {}
 		    Url_Dispatch $sock
 	        } else {
 		    Httpd_Error $sock 411 "Confusing mime headers"
@@ -314,17 +321,85 @@ proc HttpdRead {sock} {
 	    }
 	}
     } elseif {![eof $sock]} {
-	# Use counted mode to get the post data
 
-        if {[HttpdPostData $sock [read $sock $data(count)]]} {
-	    Url_Dispatch $sock
-	}
+	# TclHttpd used to read all the post data here and
+	# blindly merge it with the query data from the URL.
+	# For compatibility, this is postponed until either
+	# Url_DecodeQuery is called or Httpd_GetPostData is called.
+
+	fileevent $sock readable {}
+	Url_PostHook $sock $data(count)
+	Url_Dispatch $sock
     } else {
 	Httpd_Log $sock Error "Broken connection reading POST data"
 	Httpd_SockClose $sock 1 "broken connection during post data"
     }
 }
 
+# Httpd_PostDataSize --
+#
+# Arguments:
+#	sock	Client connection
+#
+# Results:
+#	The amount of post data available.
+
+proc Httpd_PostDataSize {sock} {
+    upvar #0 Httpd$sock data
+
+    if {![info exist data(count)]} {
+	return 0
+    }
+    return $data(count)
+}
+
+# Httpd_GetPostData --
+#
+# Arguments:
+#	sock	Client connection
+#	varName	Name of buffer variable to append post data to
+#	size	Amount of data to read this call. -1 to read all available.
+#
+# Results:
+#	The amount of data left to read.  When this goes to zero, you are done.
+
+proc Httpd_GetPostData {sock varName {size -1}} {
+    upvar #0 Httpd$sock data
+    upvar 1 $varName buffer
+
+    if {![info exist data(count)] || $data(count) == 0} {
+	return 0
+    }
+    if {$size < 0 || $size > $data(count)} {
+	set size $data(count)
+    }
+    set block [read $sock $size]
+    append buffer $block
+    set data(count) [expr {$data(count) - [string length $block]}]
+    return $data(count)
+}
+
+# Httpd_GetPostChannel --
+#
+# Arguments:
+#	sock		Client connection
+#	sizeName	Name of variable to get the amount of post
+#			data expected to be read from the channel
+#
+# Results:
+#	The socket, as long as there is POST data to read
+
+proc Httpd_GetPostChannel {sock sizeName} {
+    upvar #0 Httpd$sock data
+    upvar 1 $sizeName size
+
+    if {![info exist data(count)]} {
+	error "no post data"
+    }
+    set size $data(count)
+    return $sock
+}
+
 # HttpdPostData --
 # This accepts the post data.  There may be a conflict between
 # query data in the url and post data - you choose:
@@ -402,13 +477,21 @@ proc HttpdClose {sock} {
     return $close
 }
 
-# Utility routine for outputting response headers for normal data
-# Does not output the end of header markers so additional header lines can be added
-#  sock:  The connection handle
-#  type:  The mime type of this response
-#  close: If true, signal connection close headers.  See HttpdClose
-#  size:  The size "in bytes" of the response
-#  code:   The return code - defualts to 200
+# HttpdRespondHeader
+#
+#	Utility routine for outputting response headers for normal data Does
+#	not output the end of header markers so additional header lines can be
+#	added
+#
+# Arguments:
+#	sock	The connection handle
+#	type	The mime type of this response
+#	close	If true, signal connection close headers.  See HttpdClose
+#	size	The size "in bytes" of the response
+#	code	The return code - defualts to 200
+#
+# Side Effects:
+# 	Outputs header lines
 
 proc HttpdRespondHeader {sock type close size {code 200}} {
     global Httpd Httpd_Errors
@@ -425,14 +508,52 @@ proc HttpdRespondHeader {sock type close size {code 200}} {
 	append reply "Connection: Keep-Alive" \n
     }
     append reply "Content-Type: $type" \n
-    append reply "Content-Length: $size" \n
+    if {[string length $size]} {
+	append reply "Content-Length: $size" \n
+    }
     puts -nonewline $sock $reply
-
 }
 
-# Httpd_ReturnFile - return a file.
-# type - is a Content-Type
-# path - is the file pathname
+# Httpd_SetCookie
+#	Define a cookie to be used in a reply
+#	Call this before using Httpd_ReturnFile or
+#	Httpd_ReturnData
+#
+# Arguments:
+#	sock	handle on the connection
+#	cookie	Set-Cookie line
+
+proc Httpd_SetCookie {sock cookie} {
+    upvar #0 Httpd$sock data
+    lappend data(set-cookie) $cookie
+}
+
+# HttpdSetCookie
+#	Generate the Set-Cookie headers in a reply
+#	Use Httpd_SetCookie to register cookes eariler
+#
+# Arguments:
+#	sock	handle on the connection
+
+proc HttpdSetCookie {sock} {
+    upvar #0 Httpd$sock data
+    if {[info exist data(set-cookie)]} {
+	foreach item $data(set-cookie) {
+	    puts $sock "Set-Cookie: $item"
+	}
+    }
+}
+
+# Httpd_ReturnFile
+#	Return a file.
+#
+# Arguments:
+#	sock	handle on the connection
+#	type	is a Content-Type
+#	path	is the file pathname
+#
+# Side Effects:
+#	Sends the file contents back as the reply.
 
 proc Httpd_ReturnFile {sock type path} {
     global Httpd
@@ -442,6 +563,7 @@ proc Httpd_ReturnFile {sock type path} {
     set data(file_size) [file size $path]
     set close [HttpdClose $sock]
     HttpdRespondHeader $sock $type $close $data(file_size) 200
+    HttpdSetCookie $sock
     puts $sock "Last-Modified: [HttpdDate [file mtime $path]]"
     puts $sock ""
     if {$data(proto) != "HEAD"} {
@@ -474,6 +596,7 @@ proc Httpd_ReturnData {sock type content {code 200}} {
     Count urlreply
     set close [HttpdClose $sock]
     HttpdRespondHeader $sock $type $close [string length $content] $code
+    HttpdSetCookie $sock
     puts $sock ""
     if {$data(proto) != "HEAD"} {
 	fconfigure $sock -translation binary -blocking $Httpd(sockblock)
@@ -495,6 +618,7 @@ proc Httpd_ReturnCacheableData {sock type content date {code 200}} {
     Count urlreply
     set close [HttpdClose $sock]
     HttpdRespondHeader $sock $type $close [string length $content] $code
+    HttpdSetCookie $sock
     puts $sock "Last-Modified: [HttpdDate $date]"
     puts $sock ""
     if {$data(proto) != "HEAD"} {
@@ -614,9 +738,12 @@ proc Httpd_Redirect {newurl sock} {
     set message [format $HttpdRedirectFormat $newurl]
     set close [HttpdClose $sock]
     HttpdRespondHeader $sock text/html $close [string length $message] 302
+    HttpdSetCookie $sock
     puts $sock "Location: $newurl"
     puts $sock ""
-    puts $sock $message
+    # The -nonewline is important here to work properly with
+    # keep-alive connections
+    puts -nonewline $sock $message
     Httpd_SockClose $sock $close
 }
 
@@ -678,14 +805,14 @@ proc Httpd_RequestAuth {sock type realm} {
     HttpdRespondHeader $sock text/html $close [string length $HttpdAuthorizationFormat] 401
     puts $sock "Www-Authenticate: $type realm=\"$realm\""
     puts $sock ""
-    puts $sock $HttpdAuthorizationFormat
+    puts -nonewline $sock $HttpdAuthorizationFormat
     Httpd_SockClose $sock $close
 }
 
 # generate a date string in HTTP format
 
 proc HttpdDate {seconds} {
-    return [clock format $seconds -format {%a, %d %b %Y %T %Z} -gmt true]
+    return [clock format $seconds -format {%a, %d %b %Y %T GMT} -gmt true]
 }
 
 
